--[[
 * libsota-commands.lua
 * Copyright (C) 2022 Garritt Grandberg (aka Tirrag/peace-n-love)

 This file is a port of the commands functionality from libsota put into a module format.

   libsota is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   libsota is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.";
]]


-- TODO: validate for SES

local libsota_commands = {
}

libsota_commands.command = {
	list = {},
	add = function(command, callback)
		libsota_commands.command.list[tostring(command)] = {
			_command = command,
			callback = callback,
			shortcut = nil,
			channel = nil,
			sender = nil,
			receiver = nil,
		}
		return command
	end,
	remove = function(command)
		if libsota_commands.command.list[command].shortcut then
			libsota_commands.shortcut.remove(libsota_commands.command.list[command].shortcut)
		end
		libsota_commands.command.list[command] = nil
	end,
	invoke = function(command, ...)
		if ... then
			libsota_commands.command.list[command].callback(nil, unpack(...))
		else
			libsota_commands.command.list[command].callback()
		end
	end,
	restrict = function(command, channel, sender, receiver)
		if libsota_commands.command.list[command] then
			libsota_commands.command.list[command].channel = channel
			libsota_commands.command.list[command].sender = sender
			libsota_commands.command.list[command].receiver = receiver
		end
	end
}

libsota_commands.shortcut = {
	list = { pressed = {}, watch = {} },
	add = function(...)
		local action = "pressed"
		local keys = {...}
		local callback = keys[#keys]; keys[#keys] = nil
		local key = keys[#keys]; keys[#keys] = nil
		if not libsota_commands.shortcut.list[action][tostring(key)] then
			libsota_commands.shortcut.list[action][tostring(key)] = {}
		end
		if type(callback) == "string" then
			callback = libsota_commands.command.list[callback].callback
		end
		local id = #libsota_commands.shortcut.list[action][key] + 1
		local index = { action = action, key = key, id = id }
		libsota_commands.shortcut.list[action][key][id] = {
			_index = index,
			key = key,
			keysHeld = keys,
			callback = callback,
		}
		if type(callback) == string then
			if libsota_commands.command.list[callback].shortcut then
				libsota_commands.shortcut.remove(libsota_commands.command.list[callback].shortcut)
			end
			libsota_commands.command.list[callback].shortcut = index
		end
		return index
	end,
	remove = function(index)
		libsota_commands.shortcut.list[index.action][index.key][index.id] = nil
		if #libsota_commands.shortcut.list[index.action][index.key] == 0 then libsota_commands.shortcut.list[index.action][index.key] = nil end
	end,
	invoke = function(index)
		libsota_commands.shortcut.list[index.action][index.key][index.id].callback()
	end,
}

local command_callback_queue = {}

local function command_queue(callback, ...)
	local index = #command_callback_queue + 1

	if type(callback) == "function" then
		command_callback_queue[index] = {
			callback = callback,
			userdata = table.pack(...),
		}	
	end
end

local command_timer_internal_last_ts = nil

local function command_timer_internal()
	local ts = os.time()

	-- see if we need to run yet
	if (command_timer_internal_last_ts != nil and (ts - command_timer_internal_last_ts) < 0.01) then return end

	-- queued callbacks
	for i,q in next, command_callback_queue do
		q.callback(unpack(q.userdata))
		command_callback_queue[i] = nil
	end
	
	command_timer_internal_last_ts = ts
end


libsota_commands.onUpdate = function()
	command_timer_internal()
	
	-- check key up
	for ku,r in next, libsota_commands.shortcut.list.pressed do
		if ShroudGetOnKeyUp(ku) then
			for _,f in next, r do
				local invoke = true
				for _,kd in next, f.keysHeld do
					invoke = invoke and ShroudGetKeyDown(kd)
				end
				if invoke then f.callback() end
			end
		end
	end
	
	-- check key held/repeat
	for ku,r in next, libsota_commands.shortcut.list.watch do
		if ShroudGetOnKeyUp(ku) then
			for _,f in next, r do
				local invoke = true
				for _,kd in next, f.keysHeld do
					invoke = invoke and ShroudGetKeyDown(kd)
				end
				if invoke then
					f.timeUp = os.time()
					f.callback("up", ku, f.keysHeld)
					if os.time() - f.timeDown <= 0.25 then
						f.callback("pressed", ku, f.keysHeld)
					end
				end
			end
		end
	end
end

libsota_commands.onConsoleInput = function(channel, sender, message)

	command_queue(function(...)

		-- handle player flag changes
		if (channel == "Story" or channel == "System") and message:find("PvP") then ShroudGetPlayerName() end

		-- parse message
		local src, dst, msg = message:match("^(.-) to (.-) %[.-:%s*(.*)$")
		if sender == "" then sender = src end
		if sender == "" then sender = ShroudGetPlayerName() end
		if msg:byte() == 92 or msg:byte() == 95 or msg:byte() == 33 then
			local cmd, tail = msg:match("^[\\_!](%w+)%s*(.*)$")
			local source = {
				channel = channel,
				sender = sender,
				receiver = dst,
			}
			if libsota_commands.command.list[cmd] then			
				if not libsota_commands.command.list[cmd].sender or libsota_commands.command.list[cmd].sender == sender
				and not libsota_commands.command.list[cmd].channel or libsota_commands.command.list[cmd].channel == channel
				and not libsota_commands.command.list[cmd].receiver or libsota_commands.command.list[cmd].receiver == dst
				then
					local arg = {}
					for a in tail:gmatch("%S+") do arg[#arg + 1] = a end
					arg.n = #arg
					libsota_commands.command.list[cmd].callback(source, unpack(arg))
				end
			end
		end

	end, channel, sender, message)
end

libsota_commands.onStart = function()
	command_timer_internal()
end

function libsota_commands:initialize()
	SES.on("ShroudOnStart", libsota_commands.onStart)
	SES.on("ShroudOnUpdate", libsota_commands.onUpdate)
	SES.on("ShroudOnConsoleInput", libsota_commands.onConsoleInput)
	
	return self
end

return libsota_commands:initialize()